from typing import List, Optional

import numpy as np
import petab.C as ptc
import pyqtgraph as pg
from PySide2 import QtCore

from . import plot_row

class DottedLine:
    """
    Class for plotting lines with points and error bars.
    """
    def __init__(self):
        """
        Attributes:
            lines: List of PlotDataItems
            error_bars: List of ErrorBarItems
            p_row: PlotRow
            dataset_id: DatasetId
            is_simulation: Boolean
        """

        self.lines: List[pg.PlotDataItems] = []
        self.error_bars: List[pg.ErrorBarItems] = []

        self.p_row: Optional[plot_row.PlotRow] = None
        self.dataset_id: str = ""
        self.is_simulation: bool = False
        self.color: str = "k"
        self.style: QtCore.Qt.PenStyle = QtCore.Qt.DashDotLine

    def initialize_from_plot_row(self, p_row: plot_row.PlotRow):
        """
        Initialize all attributes with the information
        of the provided PlotRow.

        Arguments:
            p_row: The PlotRow containing the plotting
                    specifications.
        """
        self.p_row = p_row
        self.dataset_id = p_row.dataset_id
        self.is_simulation = p_row.is_simulation
        if self.is_simulation:
            self.dataset_id += "_simulation"
            self.style = QtCore.Qt.SolidLine

        self.generate_line()

        # Only add error bars when needed
        if (self.p_row.has_replicates or
            self.p_row.plot_type_data == ptc.PROVIDED) \
                and self.p_row.plot_type_data != ptc.REPLICATE:
            self.add_error_bars()

    def initialize(self, lines, error_bars, dataset_id, is_simulation):
        """
        Initialize the attributes equal to the given arguments.
        """
        self.lines = lines
        self.error_bars = error_bars
        self.dataset_id = dataset_id
        self.is_simulation = is_simulation
        if self.is_simulation:
            self.dataset_id += "_simulation"
            self.style = QtCore.Qt.SolidLine

    def generate_line(self):
        """
        Create a PlotDataItem for the line
        which will be stored in self.lines
        The list will only contain more than one object
        when plotting replicates.
        """
        legend_name = self.p_row.legend_name
        symbol = "o"
        if self.is_simulation:
            legend_name = legend_name + " simulation"
            symbol = "t"

        if self.p_row.plot_type_data == ptc.REPLICATE:
            x_data = self.p_row.get_replicate_x_data()
            y_data = self.p_row.get_replicate_y_data()
            first_replicate = True
            for x, y in zip(x_data, y_data):
                if first_replicate:
                    self.lines.append(pg.PlotDataItem(x, y,
                                                      name=legend_name,
                                                      symbolPen=pg.mkPen("k"),
                                                      symbol=symbol,
                                                      symbolSize=7)
                                                      )
                    first_replicate = False
                else:
                    # if all would replicate have a legend_name,
                    # that name would be duplicated in the legend
                    self.lines.append(pg.PlotDataItem(x, y))
        else:
            self.lines.append(pg.PlotDataItem(self.p_row.x_data,
                                              self.p_row.y_data,
                                              name=legend_name,
                                              symbolPen=pg.mkPen("k"),
                                              symbol=symbol, symbolSize=7))

    def add_error_bars(self):
        """
        Create an ErrorBarItem based on the information
        of the PlotRow.
        """
        error_length = self.p_row.sd
        if self.p_row.plot_type_data == ptc.MEAN_AND_SEM:
            error_length = self.p_row.sem
        elif self.p_row.plot_type_data == ptc.PROVIDED:
            error_length = self.p_row.provided_noise
        beam_width = 0
        if len(self.p_row.x_data) > 0:  # self.p_row.x_data could be empty
            beam_width = (np.max(self.p_row.x_data) -
                          np.min(self.p_row.x_data)) / 100
        error = pg.ErrorBarItem(x=self.p_row.x_data, y=self.p_row.y_data,
                                top=error_length, bottom=error_length,
                                beam=beam_width)
        self.error_bars.append(error)

    def add_to_plot(self, plot, color="k", add_error_bars=True):
        """
        Add all lines and error bars of
        this object to the provided plot.
        The color of the lines and the points can
        also be provided.

        Arguments:
            plot: The plot to which everything
                    should be added.
            color: The color the line and the
                    points should have.
        """
        self.color = color

        for line in self.lines:
            line.setPen(color, style=self.style, width=2)
            line.setSymbolBrush(color)
        for error_bars in self.error_bars:
            error_bars.setData(pen=pg.mkPen(color))
        self.enable_in_plot(plot, add_error_bars)

    def enable_in_plot(self, plot, add_error_bars=True):
        """
        Add all lines and error bars of
        this object to the provided plot.
        """
        for line in self.lines:
            plot.addItem(line)
        if add_error_bars:
            for error_bars in self.error_bars:
                plot.addItem(error_bars)

    def disable_in_plot(self, plot):
        """
        Remove all lines and error bars of
        this object to the provided plot.
        """
        for line in self.lines:
            plot.removeItem(line)
        for error_bars in self.error_bars:
            plot.removeItem(error_bars)

    def hide_lines(self):
        for line in self.lines:
            line.setPen(None)

    def hide_points(self):
        for line in self.lines:
            line.setSymbolPen(None)
            line.setSymbolBrush(None)

    def show_lines(self):
        for line in self.lines:
            line.setPen(self.color, style=self.style, width=2)

    def show_points(self):
        for line in self.lines:
            line.setSymbolBrush(self.color)
            line.setSymbolPen("k")



import os
import sys
from pathlib import Path

import pandas as pd
import petab
import petab.C as ptc
from PySide2 import QtWidgets, QtCore, QtGui
from PySide2.QtCore import (Qt, QSortFilterProxyModel)
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import (QAction, QVBoxLayout, QHeaderView,
                               QSizePolicy, QTableView, QWidget, QFileDialog)
from petab import core
from petab.visualize.helper_functions import check_ex_exp_columns

from . import table_models


class TableWidget(QWidget):
    """Widget for displaying a PEtab table."""

    def __init__(self, data: pd.DataFrame, add_checkbox_col: bool, window):
        QWidget.__init__(self)
        self.window = window

        # Set the Model
        if add_checkbox_col:
            self.model = table_models.VisualizationTableModel(data, window)
        elif window.exp_data.equals(data):  # for any other df
            self.model = table_models.MeasurementTableModel(data, window)
        else:
            self.model = table_models.PetabTableModel(data)

        # Creating a QTableView
        self.table_view = QTableView()
        self.filter_proxy = QSortFilterProxyModel()
        self.filter_proxy.setSourceModel(self.model)
        self.table_view.setModel(self.filter_proxy)
        self.table_view.setSortingEnabled(True)

        # add a checkbox column for visualization dfs
        if add_checkbox_col:
            delegate = table_models.CheckBoxDelegate(None)
            self.table_view.setItemDelegateForColumn(0, delegate)

        # QTableView Headers
        self.horizontal_header = self.table_view.horizontalHeader()
        self.horizontal_header.setSortIndicator(-1, Qt.DescendingOrder)
        self.vertical_header = self.table_view.verticalHeader()
        self.horizontal_header.setSectionResizeMode(
            QHeaderView.ResizeToContents)

        # QWidget Layout
        self.main_layout = QVBoxLayout()
        size = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)

        size.setHorizontalStretch(1)
        self.table_view.setSizePolicy(size)
        self.main_layout.addWidget(self.table_view)

        self.setLayout(self.main_layout)

    def closeEvent(self, event):
        if self in self.window.popup_windows:
            self.window.popup_windows.remove(self)
        super().closeEvent(event)


def pop_up_table_view(window: QtWidgets.QMainWindow, df: pd.DataFrame):
    """
    Create a popup window that displays the dataframe.

    Arguments:
        window: The main window to which the TableWidget gets added.
        df: The dataframe to display
    """
    add_checkbox_col = False
    if window.visualization_df is not None\
            and window.visualization_df.equals(df):
        add_checkbox_col = True
    popup_window = TableWidget(data=df,
                               add_checkbox_col=add_checkbox_col,
                               window=window)
    popup_window.setGeometry(QtCore.QRect(100, 100, 800, 400))
    popup_window.show()
    window.popup_windows.append(popup_window)


def table_tree_view(window: QtWidgets.QMainWindow, folder_path):
    """
    Create a treeview of the yaml file.
    Set the windows df attributes equal to the first file
    in each branch.

    Arguments:
        window: The main window to which the treeview is added
        folder_path: The path to the folder the yaml file is in
    """
    model = QtGui.QStandardItemModel()
    tree_view = window.tree_view
    root_node = model.invisibleRootItem()

    tidy_names = {ptc.MEASUREMENT_FILES: "Measurement Tables",
                  ptc.VISUALIZATION_FILES: "Visualization Tables",
                  ptc.CONDITION_FILES: "Condition Tables",
                  ptc.OBSERVABLE_FILES: "Observable Tables",
                  ptc.SBML_FILES: "SBML Files"}

    # iterate through the yaml_dict
    for key in window.yaml_dict:
        branch = QtGui.QStandardItem(tidy_names[key])
        branch.setEditable(False)
        is_first_df = True

        # iterate through the files of a yaml_dict entry
        for filename in window.yaml_dict[key]:
            file = QtGui.QStandardItem(filename)
            df = None
            if key == ptc.MEASUREMENT_FILES:
                df = petab.get_measurement_df(folder_path + "/" + filename)
                if is_first_df:
                    window.exp_data = df
            if key == ptc.VISUALIZATION_FILES:
                df = petab.get_visualization_df(folder_path + "/" + filename)
                df.insert(0, "Displayed", 1)  # needed for the checkbox column
                if is_first_df:
                    window.visualization_df = df
            if key == ptc.CONDITION_FILES:
                df = petab.get_condition_df(folder_path + "/" + filename)
                if is_first_df:
                    window.condition_df = df
            if key == ptc.OBSERVABLE_FILES:
                df = petab.get_observable_df(folder_path + "/" + filename)
                if is_first_df:
                    window.observable_df = df
            file.setData(df, role=Qt.UserRole + 1)
            branch.appendRow(file)
            is_first_df = False
        root_node.appendRow(branch)

    if window.simulation_df is not None:
        branch = QtGui.QStandardItem("simulation_files")
        simulation_file = QtGui.QStandardItem("simulation_file")
        simulation_file.setData(window.simulation_df, role=Qt.UserRole + 1)
        branch.appendRow(simulation_file)
        root_node.appendRow(branch)

    tree_view.setModel(model)
    reconnect(tree_view.clicked,
              lambda i: exchange_dataframe_on_click(i, model,
                                                    window, tidy_names))
    reconnect(tree_view.doubleClicked,
              lambda i: display_table_on_doubleclick(i, model, window))


def reconnect(signal, new_function=None):
    """
    Disconnect a signal from all functions and connect it
    to a new function.

    Arguments:
        signal: The signal to reconnect
        new_function: The function the signal gets connected to
    """
    try:
        signal.disconnect()
    except RuntimeError:
        pass
    if new_function is not None:
        signal.connect(new_function)


def exchange_dataframe_on_click(index: QtCore.QModelIndex,
                                model: QtGui.QStandardItemModel,
                                window: QtWidgets.QMainWindow,
                                tidy_names: dict):
    """
    Changes the currently plotted dataframe with the one that gets clicked on
    and replot the data, e.g. switch the measurement or visualization df.

    Arguments:
        index: index of the clicked dataframe
        model: model containing the data
        window: Mainwindow whose attributes get updated
    """
    df = model.data(index, role=Qt.UserRole + 1)
    parent = index.parent()
    parent_name = model.data(parent, QtCore.Qt.DisplayRole)
    # Only replot when a new dataframe is selected
    # (Important because double clicking also calls this function)
    df_changed = True
    if df is None:
        return
    if parent_name == tidy_names[ptc.MEASUREMENT_FILES]:
        if window.exp_data.equals(df):
            df_changed = False
        window.exp_data = df
    if parent_name == tidy_names[ptc.VISUALIZATION_FILES]:
        if window.visualization_df.equals(df):
            df_changed = False
        window.visualization_df = df
    if parent_name == tidy_names[ptc.CONDITION_FILES]:
        if window.condition_df.equals(df):
            df_changed = False
        window.condition_df = df
    if parent_name == tidy_names[ptc.OBSERVABLE_FILES]:
        if window.observable_df.equals(df):
            df_changed = False
        window.observable_df = df
    if parent_name == "simulation_files":
        if window.simulation_df.equals(df):
            df_changed = False
        window.simulation_df = df

    if df_changed:
        window.add_plots()


def display_table_on_doubleclick(index: QtCore.QModelIndex,
                                 model: QtGui.QStandardItemModel,
                                 window: QtWidgets.QMainWindow):
    """
    Display the dataframe in a new window upon double-click.

    Arguments:
        index: index of the clicked dataframe
        model: model containing the data
        window: Mainwindow whose attributes get updated
    """
    df = model.data(index, role=Qt.UserRole + 1)
    if df is not None:
        pop_up_table_view(window, df)


def add_file_selector(window: QtWidgets.QMainWindow):
    """
    Add a file selector button to the main window.
    Arguments:
        window: Mainwindow
    """
    open_yaml_file = QAction(QIcon('open.png'), 'Open YAML file...', window)
    open_yaml_file.triggered.connect(lambda x: show_yaml_dialog(window))
    open_simulation_file = QAction(QIcon('open.png'),
                                   'Open simulation file...', window)
    open_simulation_file.triggered.connect(
        lambda x: show_simulation_dialog(window))
    quit = QAction("Quit", window)
    quit.triggered.connect(sys.exit)

    menubar = window.menuBar()
    file_menu = menubar.addMenu('&Select File')
    file_menu.addAction(open_yaml_file)
    file_menu.addAction(open_simulation_file)
    file_menu.addAction(quit)


def show_yaml_dialog(window: QtWidgets.QMainWindow):
    """
    Display a file selector window when clicking on the select YAML file menu
    item, then display the new plots described by the YAML file.

    Arguments:
        window: Mainwindow
    """
    home_dir = str(Path.home())
    # start file selector on the last selected directory
    settings = QtCore.QSettings("petab", "petabvis")
    if settings.value("last_dir") is not None:
        home_dir = settings.value("last_dir")
    file_name = QFileDialog.getOpenFileName(window, 'Open file', home_dir)[0]
    if file_name != "":  # if a file was selected
        # save the directory for the next use
        last_dir = os.path.dirname(file_name) + "/"
        settings.setValue("last_dir", last_dir)

        window.warn_msg.setText("")
        window.warnings.clear()
        window.warning_counter.clear()

        # select the first df in the dict for measurements, etc.
        yaml_dict = petab.load_yaml(file_name)["problems"][0]
        window.yaml_dict = yaml_dict
        if ptc.VISUALIZATION_FILES not in yaml_dict:
            window.visualization_df = None
            window.add_warning(
                "The YAML file contains no "
                "visualization file (default plotted)")
        window.simulation_df = None

        # table_tree_view sets the df attributes of the window
        # equal to the first file of each branch
        # (measurement, visualization, ...)
        table_tree_view(window, last_dir)
        window.add_plots()


def show_simulation_dialog(window: QtWidgets.QMainWindow):
    """
    Display a file selector window when clicking
    on the select simulation file button,
    then add the simulation lines to the plots.

    Arguments:
        window: main window.
    """
    home_dir = str(Path.home())
    settings = QtCore.QSettings("petab", "petabvis")
    if settings.value("last_dir") is not None:
        home_dir = settings.value("last_dir")
    file_name = QFileDialog.getOpenFileName(
        window, 'Open simulation file', home_dir)[0]
    if file_name:  # if a file was selected
        if window.exp_data is None:
            window.add_warning("Please open a YAML file first.")
        else:
            window.warn_msg.setText("")
            window.warnings.clear()
            window.warning_counter.clear()

            sim_data = core.get_simulation_df(file_name)
            # check columns, and add non-mandatory default columns
            sim_data, _, _ = check_ex_exp_columns(
                sim_data, None, None, None, None, None,
                window.condition_df, sim=True)
            # delete the replicateId column if it gets added to the simulation
            # table but is not in exp_data because it causes problems when
            # splitting the replicates
            if ptc.REPLICATE_ID not in window.exp_data.columns \
                    and ptc.REPLICATE_ID in sim_data.columns:
                sim_data.drop(ptc.REPLICATE_ID, axis=1, inplace=True)

            if len(window.yaml_dict[ptc.MEASUREMENT_FILES]) > 1:
                window.add_warning(
                    "Not Implemented Error: Loading a simulation file with "
                    "multiple measurement files is currently not supported.")
            else:
                window.simulation_df = sim_data
                window.add_plots()

                # insert correlation plot at position 1
                window.wid.insertWidget(1, window.plot2_widget)
                table_tree_view(window, os.path.dirname(file_name))

        # save the directory for the next use
        last_dir = os.path.dirname(file_name) + "/"
        settings.setValue("last_dir", last_dir)



import numpy as np
import pandas as pd
import petab.C as ptc

from . import row_class
from . import utils


class PlotRow(row_class.RowClass):
    """
    Can add the content of a visualization_df row to a PlotItem.
    Used for line plots.

    Attributes:
        x_data: Numpy array of the x-values
        y_data: Numpy array of the y-values
        sd: Standard deviation of the replicates
        sem: Standard error of the mean of the replicates
        provided noise: Noise of the measurements
    """

    def __init__(self, exp_data: pd.DataFrame,
                 plot_spec: pd.Series, condition_df: pd.DataFrame, ):

        super().__init__(exp_data, plot_spec, condition_df)

        # calculate new attributes
        self.y_data = self.get_y_data()
        self.x_data = self.get_x_data()
        self.sd = utils.sd_replicates(self.line_data, self.x_var,
                                      self.is_simulation)
        self.sem = utils.sem_replicates(self.line_data, self.x_var,
                                        self.is_simulation)
        self.provided_noise = self.get_provided_noise()

    def get_x_data(self):
        """
        Return the x-values that should be plotted.

        Returns:
            x_data: The x-values as numpy array
        """
        # for replicate plots
        if self.plot_type_data == ptc.REPLICATE\
                and ptc.REPLICATE_ID in self.line_data.columns:
            return np.hstack(self.get_replicate_x_data())

        # for concentration plots
        if self.x_var != ptc.TIME:
            x_data = self.condition_df[self.x_var]

        else:  # for time plots
            if self.has_replicates:
                # to keep the order intact
                # (only needed if no replicate id col is provided)
                x_data = np.array(sorted(set(self.replicates[0][self.x_var])))
            else:
                x_data = np.asarray(self.replicates[0][self.x_var])
            x_data = x_data + self.x_offset

        return x_data

    def get_y_data(self):
        """
        Return the mean of the y-values that should be plotted if
        the plot type is not ptc.REPLICATE or there was no replicateId
        provided (then the mean will still be plotted together with the
        min and max values of the replicates).
        Otherwise, if the replicateId is provided,
        return the y-values of all replicates.


        Returns:
            y_data: The y-values as numpy array
        """
        # for replicate plots
        if self.plot_type_data == ptc.REPLICATE:
            return np.hstack(self.get_replicate_y_data())

        # variable is either measurement or simulation
        variable = self.get_y_variable_name()
        y_data = utils.mean_replicates(self.line_data, self.x_var,
                                       variable)
        y_data = y_data + self.y_offset

        return y_data

    def get_replicate_x_data(self):
        """
        Return the x-values of each replicate as a list of lists

        Returns:
            x_data_replicates: The y-values for each replicate
        """
        x_data = []

        if self.x_var != ptc.TIME:
            x_values = np.asarray(self.condition_df[self.x_var])
            x_values = x_values + self.x_offset
            # the x-values are the same for replicates of
            # concentration plots, thus we repeat them until
            # the x- and y-values have the same length
            if len(x_values) != len(self.y_data)\
                    and len(self.y_data) % len(x_values) == 0:
                x_data = [x_values for _
                          in range(int(len(self.y_data)/len(x_values)))]
                return x_data
            else:
                raise Exception("Error occurred when deriving the x-values "
                                "for replicates of a concentration plot")

        default_x_values = [x for _, x in
                            self.replicates[0].groupby(self.x_var, sort=True)]
        for replicate in self.replicates:
            if ptc.REPLICATE_ID in self.line_data.columns:
                x_values = np.asarray(replicate[self.x_var])
            else:
                # when no explicit replicate id is given, we assume that
                # each replicate uses the same x-values which are determined
                # by the unique x-values in the data
                x_values = default_x_values
            x_values = x_values + self.x_offset
            x_data.append(x_values)

        return x_data

    def get_replicate_y_data(self):
        """
        Return the y-values of each replicate as a list of lists

        Returns:
            y_data_replicates: The y-values for each replicate
        """
        y_data = []
        # variable is either measurement or simulation
        variable = self.get_y_variable_name()
        for replicate in self.replicates:
            y_values = utils.mean_replicates(replicate, self.x_var,
                                             variable)
            y_values += self.y_offset
            y_data.append(y_values)
        return y_data



import petab.C as ptc
from PySide2 import QtWidgets, QtCore, QtGui
from PySide2.QtCore import (QAbstractTableModel, QModelIndex, Qt)
from PySide2.QtGui import QColor


class PetabTableModel(QAbstractTableModel):
    """PEtab data table model."""

    def __init__(self, df=None):
        QAbstractTableModel.__init__(self)
        self.load_data(df)
        self.df = df
        self.row_count = df.shape[0]
        self.column_count = df.shape[1]

    def load_data(self, data):
        for x in data:
            setattr(self, x, data[x])

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        row = index.row()
        column = index.column()
        self.df.iloc[row, column] = value
        return True

    def rowCount(self, parent=QModelIndex()):
        return self.row_count

    def columnCount(self, parent=QModelIndex()):
        return self.column_count

    def headerData(self, section, orientation, role=None):
        if role != Qt.DisplayRole:
            return None

        if orientation == Qt.Horizontal:
            return self.df.columns[section]
        else:
            return self.df.index[section]

    def data(self, index, role=Qt.DisplayRole):
        column = index.column()
        row = index.row()
        if role == Qt.DisplayRole:
            return str(self.df.iloc[row, column])

        elif role == Qt.BackgroundRole:
            return QColor(Qt.white)

        elif role == Qt.TextAlignmentRole:
            return Qt.AlignRight

        return None

    def get_value(self, row, column):
        return self.df.iloc[row][column]


class VisualizationTableModel(PetabTableModel):
    """
    Special table model for visualization files.

    Make the first column of the table editable for the checkbox column.
    Highlight the rows of the currently displayed plot.
    """

    def __init__(self, df=None, window=None):
        PetabTableModel.__init__(self, df)
        self.window = window

    def flags(self, index):
        if not index.isValid():
            return 0

        if index.column() == 0:
            return Qt.ItemIsEditable | Qt.ItemIsEnabled \
                   | Qt.ItemIsSelectable | Qt.ItemIsUserCheckable

        return Qt.ItemIsSelectable | Qt.ItemIsEnabled

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.BackgroundRole:
            current_plot = self.window.vis_spec_plots[
                self.window.current_list_index]
            current_plot_id = current_plot.plot_id
            if self.df[ptc.PLOT_ID][index.row()] == current_plot_id:
                return QtGui.QColor("yellow")
        else:
            return super().data(index, role)

    def get_window(self):
        return self.window


class MeasurementTableModel(PetabTableModel):
    """
    Special table model for measurement files.

    Highlight the rows of the currently displayed plot.
    """
    def __init__(self, df=None, window=None):
        PetabTableModel.__init__(self, df)
        self.window = window
        self.current_plot_id = ""
        self.current_dataset_ids = []
        self.current_observable_ids = []

    def data(self, index, role=Qt.DisplayRole):
        if role != Qt.BackgroundRole:
            return super().data(index, role)

        current_plot = self.window.vis_spec_plots[
            self.window.current_list_index]
        # for default plots, plot_id is the observableId
        # otherwise it is the datasetId
        plot_id = current_plot.plot_id
        row = self.df.iloc[index.row()]

        # for default plots
        if self.window.visualization_df is None:
            if row[ptc.OBSERVABLE_ID] == plot_id:
                return QtGui.QColor("yellow")
            else:
                return super().data(index, role)

        # only recalculate the plot_ids and
        # observable ids if the current plot changes
        if plot_id != self.current_plot_id:
            vis_df = self.window.visualization_df
            if ptc.DATASET_ID in vis_df.columns:
                self.current_dataset_ids = list(vis_df[
                    vis_df[ptc.PLOT_ID] == plot_id]
                    [ptc.DATASET_ID].unique())
            if ptc.Y_VALUES in vis_df.columns:
                self.current_observable_ids = list(vis_df[
                    vis_df[ptc.PLOT_ID] == plot_id]
                    [ptc.Y_VALUES].unique())
            self.current_plot_id = plot_id

        correct_dataset_id = True
        correct_observable_id = True
        if self.current_dataset_ids and ptc.DATASET_ID in row.index:
            disabled_ids = current_plot.disabled_rows
            correct_dataset_id = row[ptc.DATASET_ID] in \
                self.current_dataset_ids and row[ptc.DATASET_ID] \
                not in disabled_ids
        if self.current_observable_ids:
            correct_observable_id = row[ptc.OBSERVABLE_ID] \
                                    in self.current_observable_ids
        if correct_dataset_id and correct_observable_id:
            return QtGui.QColor("yellow")

        return super().data(index, role)

    def get_window(self):
        return self.window


class CheckBoxDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning QCheckBox cell to the column to
    which it is applied.

    Used for the visualization table to add the checkbox column and provide
    its functionality.
    """

    def __init__(self, parent):
        QtWidgets.QItemDelegate.__init__(self, parent)

    def createEditor(self, parent, option, index):
        """
        Important, otherwise an editor is created if the user clicks in this
        cell.
        """
        return None

    def paint(self, painter, option, index):
        """
        Paint a checkbox without the label.
        """
        self.drawCheck(painter, option, option.rect,
                       QtCore.Qt.Unchecked if int(
                        index.data()) == 0 else QtCore.Qt.Checked)

    def editorEvent(self, event, model, option, index):
        """
        Change the data in the model and the state of the checkbox
        if the user presses the left mousebutton and this cell is editable.
        Otherwise do nothing.
        """
        if not int(index.flags() & QtCore.Qt.ItemIsEditable) > 0:
            return False

        if event.type() == QtCore.QEvent.MouseButtonRelease \
                and event.button() == QtCore.Qt.LeftButton:

            window = model.sourceModel().get_window()
            if ptc.DATASET_ID not in window.visualization_df.columns:
                window.add_warning("Lines without datasetId " +
                                   "can not be removed")
            # Change the checkbox-state
            plot_id = model.sourceModel().get_value(index.row(), ptc.PLOT_ID)
            dataset_id = model.sourceModel().get_value(index.row(),
                                                       ptc.DATASET_ID)
            # Set `vis_spec_plot` to the one that matches `plot_id`
            for vis_spec_plot in window.vis_spec_plots:
                if vis_spec_plot.plot_id == plot_id:
                    break
            vis_spec_plot.add_or_remove_line(dataset_id)
            self.setModelData(None, model, index)
            return True

        return False

    def setModelData(self, editor, model, index):
        """
        Change the state of the checkbox after it was clicked.
        """
        model.setData(index, 1 if int(index.data()) == 0 else 0,
                      QtCore.Qt.EditRole)



import numpy as np
import pandas as pd
import petab.C as ptc
import pyqtgraph as pg

from . import bar_row
from . import plot_class


class BarPlot(plot_class.PlotClass):
    """
     Can generate a bar plot based on the given specifications

     Arguments:
         measurement_df: PEtab measurement table
         visualization_df: PEtab visualization table
         simulation_df: PEtab simulation table
         condition_df: PEtab condition table
         plot_id: Id of the plot (has to in the visualization_df as well)

     Attributes:
         bar_rows: A list of BarRows (one for each visualization df row)
         overview_df: A df containing the information of each bar
     """

    def __init__(self, measurement_df: pd.DataFrame = None,
                 visualization_df: pd.DataFrame = None,
                 simulation_df: pd.DataFrame = None,
                 condition_df: pd.DataFrame = None,
                 plot_id: str = ""):
        super().__init__(measurement_df, visualization_df, simulation_df,
                         condition_df, plot_id)

        self.bar_width = 0.4
        # bar_rows also contains simulation bars
        self.bar_rows = []
        self.add_bar_rows(self.measurement_df)  # list of plot_rows
        self.add_bar_rows(self.simulation_df)

        # A df containing the information needed to plot the bars
        self.overview_df = pd.DataFrame(
            columns=["x", "y", "name", "sd", "sem", "provided_noise",
                     "is_simulation", "tick_pos"])

        self.plot_everything()

    def plot_everything(self):
        """
        Calculate the overview_df and create the plot based on it.
        If a simulation df is given, also generate the correlation plot.
        """
        self.plot.clear()
        self.overview_df = self.get_bars_df()

        self.generate_plot()

        if self.simulation_df is not None:
            # create correlation plot
            self.generate_correlation_plot(self.overview_df)

    def add_bar_rows(self, df):
        """
        Add a BarRow object for each row of the visualization df.

        Arguments:
            df: Measurement or Simulation df
        """
        if self.visualization_df is not None:
            for _, plot_spec in self.visualization_df.iterrows():
                if df is not None:
                    row = bar_row.BarRow(df, plot_spec, self.condition_df)
                    self.bar_rows.append(row)

    def generate_overview_df(self):
        """
        Generate the overview df containing the x- and y-data, the name,
        the dataset_id, the x_var and simulation information of all enabled
        plotRows.

        Returns:
            overview_df: A dataframe containing an overview of the plotRows
        """
        overview_df = pd.DataFrame(
            columns=["y", "name", "is_simulation", "dataset_id",
                     "sd", "sem"])
        if self.visualization_df is not None:
            dfs = [bar.get_data_df() for bar in
                   self.bar_rows
                   if bar.dataset_id not in self.disabled_rows]
            if dfs:
                overview_df = pd.concat(dfs, ignore_index=True)
        return overview_df

    def get_bars_df(self):
        """
        Generate a dataframe containing plotting information
        of the individual bars.

        Arguments:
            bar_rows: A list of BarRow objects
        Returns:
            df: A dataframe with information relevant
                for plotting a bar (x, y, sd, etc.)
        """
        df = self.generate_overview_df()

        x = list(range(len(df.index)))
        tick_pos = list(range(len(df.index)))
        for i_name, (name, name_df) in enumerate(df.groupby('name')):
            for i_replicate, i_row in enumerate(name_df.index):
                x[i_row] = i_name - np.linspace(start=0, stop=self.bar_width, num=len(name_df.index))[i_replicate]
                tick_pos[i_row] = i_name - self.bar_width / 2
        df = df.assign(x=x, tick_pos=tick_pos)

        # Adjust x and tick_pos of the bars when simulation bars are plotted
        # such that they are next to each other
        if self.simulation_df is not None:
            x = df["x"].tolist()
            tick_pos = df["tick_pos"].tolist()
            for i_name, (name, name_df) in enumerate(df.groupby('name')):
                num_replicates = len(name_df.index) / 2  # /2 due to simulation
                shift_start = self.bar_width / (2 * num_replicates)
                for i_replicate, i_row in enumerate(name_df[name_df["is_simulation"]].index):
                    tick_pos[i_row] = i_name
                    shift = np.linspace(start=shift_start,
                                        stop=self.bar_width + shift_start,
                                        num=int(num_replicates))[i_replicate]
                    x[i_row] = i_name + shift
                for i_replicate, i_row in enumerate(name_df[~name_df["is_simulation"]].index):
                    tick_pos[i_row] = i_name
                    shift = np.linspace(start=shift_start,
                                        stop=self.bar_width + shift_start,
                                        num=int(num_replicates))[i_replicate]
                    x[i_row] = i_name - shift
            df = df.assign(x=x, tick_pos=tick_pos)

        return df

    def generate_plot(self):
        """
        Generate the plot based on the information in the overview_df df.
        Add error bars to the plot.
        Set y-scale to log10 if necessary.
        """

        if not self.overview_df.empty:
            # get the axis labels info from the first line of the plot
            self.plot.setLabel("left", self.bar_rows[0].y_label)
            self.plot.setLabel("bottom", self.bar_rows[0].x_label)

            bar_width = self.bar_width
            # adjust the barwidth when plotting replicates
            if self.bar_rows[0].plot_type_data == ptc.REPLICATE:
                max_num_replicates = max(len(bar.replicates) for bar
                                         in self.bar_rows)
                bar_width = self.bar_width / max_num_replicates

            # Add bars
            simu_rows = self.overview_df["is_simulation"]
            bar_item = pg.BarGraphItem(x=self.overview_df[~simu_rows]["x"],
                                       height=self.overview_df[~simu_rows][
                                        "y"], width=bar_width,
                                       pen=pg.mkPen("b", width=2),
                                       name="measurement")
            self.plot.addItem(bar_item)  # measurement bars
            if self.simulation_df is not None:
                bar_item = pg.BarGraphItem(x=self.overview_df[simu_rows]["x"],
                                           name="simulation",
                                           height=self.overview_df[simu_rows]["y"],
                                           width=bar_width,
                                           pen=pg.mkPen("y", width=2))
                self.plot.addItem(bar_item)  # simulation bars

            # Add error bars
            error_length = self.overview_df["sd"]
            if self.bar_rows[0].plot_type_data == ptc.MEAN_AND_SEM:
                error_length = self.overview_df["sem"]
            if self.bar_rows[0].plot_type_data == ptc.PROVIDED:
                error_length = self.overview_df["provided_noise"]
            error = pg.ErrorBarItem(x=self.overview_df["x"],
                                    y=self.overview_df["y"],
                                    top=error_length, bottom=error_length,
                                    beam=bar_width/3)
            self.plot.addItem(error)

            # set tick names to the legend entry of the bars
            xax = self.plot.getAxis("bottom")
            ticks = [list(
                zip(self.overview_df["tick_pos"], self.overview_df["name"]))]
            xax.setTicks(ticks)

            # set y-scale to log if necessary
            if "log" in self.bar_rows[0].y_scale:
                self.plot.setLogMode(y=True)
                if self.plot_rows[0].x_scale == "log":
                    self.add_warning(
                        "log not supported, using log10 " +
                        "instead (in " + self.plot_title + ")")

    def add_or_remove_line(self, dataset_id):
        """
        Add the bar corresponding to the dataset id to the plot
        if it is currently disabled. Otherwise, remove it from the plot.
        Call this method when enabling/disabling rows in
        the visualization df.

        Arguments:
            dataset_id: The id of the bar which should be removed/added
        """
        if dataset_id in self.disabled_rows:
            self.disabled_rows.remove(dataset_id)
        else:
            self.disabled_rows.add(dataset_id)
        # also replots the correlation plot
        self.plot_everything()
